//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	Ds9.cc:	--- implementation of Ds9 class
//
//	CHANGE HISTORY:
//		UL, 1999-05-06: 1.0.0: - first version; uses Ds9-1.8-beta
//		UL, 1999-05-19: 1.1.0: - updated for Ds9-1.8.2-beta
//							   - set coordinate system for region markers
//								 to fk5 before reading region data
//		UL, 1999-08-17: 1.2.0: - bug fix in Ds9::regionSelection()
//		UL, 1999-09-20: 1.2.1: - do no longer check return of of 'system'
//								 used to bring up Ds9 (causes problem at
//								 Leceister)
//							   - inquire regions from Ds9 in SAOtng format
//								 for ingesiton by
//								 ExprTranslator::toSelectExpression in
//							     selectlib package
//		UL, 1999-10-27: 1.2.2: - brought in line with ds9-1.8.6.9
//								 (uses no longer '-' in front of options
//								 of XPA access points)
//		UL, 1999-10-28: 1.2.3: - in Ds9::receive(): moved first
//								 switchSignalHanlders() statement to the
//								 beginning of the function; otherwise
//								 xmgr interferes with the xmmselect-ds9
//								 pipe
//		UL, 1999-12-17: 1.3.0: - brought in line with selectlib-4.23
//		UL, 2000-01-05: 1.4.0: - only change settings (new colormap/log scale)
//								 when bringing up new ds9
//								 (fixes SPR 1034 : should not reset ds9
//								 settings)
//		UL, 2000-01-19: 1.5.0: - implemented Ds9::displayRegion()
//		UL, 2000-02-02: 1.6.0: - write "linear" keyword to Ds9 region file only
//								 if marker point coordinates are in WCS
//		UL, 2000-03-14: 1.7.0: - no longer do a "regions deleteall" before
//								 displaying a new one
//		UL, 2000-07-21: 1.8.0: - adapted Ds9::toRegion()/<<operator to
//								 ds9 region file format 3.0
//		UL, 2000-12-22: 1.8.1: - trivial change in Ds9::toRegion() to make
//								 optimizer happy
//								 
//
//	AUTHOR:
//		U. Lammers (UL), Thu May  6 11:20:00 MET DST 1999
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/task_devel/packages/imgdisplay/src/Ds9.cc,v 1.22 2000/12/22 12:23:47 ulammers Exp $
//

#include <fstream>
#include <iostream>
#include <cstdlib>
#include <cctype>
#include <cstdio>	//	tmpnam()
#include <csignal>
#include <unistd.h>
#include <streambuf.h>	//	ios::sync_with_stdio
#include "Ds9.h"
#include "Region.h"
#include "ASCregion.h"
#include "errstr.h"
#include "stringutil.h"
#include "common.h"

const string Ds9::Xpaset("xpaset");
const string Ds9::Xpaget("xpaget");
const string Ds9::TITENV("LOGNAME");


//
//	switchSignalHandlers: toggle the signal handlers for the signal SIGCLD
//						  ACEgrOpen erroneously registers a handler for
//						  the signal SIGCLD which simply exits the application.
//						  This prevents other pipes from being created and
//						  closed because pclose() emits SIGCLD to signal
//						  that the process who is reading from the pipe
//						  has passed away.
//						  This workaround can go when we switch to
//						  grace-5.0.2 in which libgrace is much cleaner
//						  implemented.
//
static void switchSignalHandlers()
{
	static void (*sigChildHandler)(int) = 0;

	if (sigChildHandler) {
		//	there is a valid signal handler - register it
		signal(SIGCLD, sigChildHandler);
		sigChildHandler = 0;
	} else
		//	there is no signal handler
		sigChildHandler = signal(SIGCLD, SIG_IGN);
} // switchSignalHandlers()


//	construct Ds9 wrapper object
Ds9::Ds9(const string &name) : ImageDisplayer(), _DS9(name)
{
	PROC

	ios::sync_with_stdio();	//	we use popen/pclose
	if (isThere()) {
		_title = "Ds9";
		char *p = getenv(TITENV.c_str());
		if (p)
			_title += "-" + string(p);

		if (!isRunning()) {
			//	is's not running - try to bring it up
			execBackground(_DS9 + " -title " + _title +
						   " -horzgraph -vertgraph");
			//	wait till it comes up and is receptive
			unsigned nChecks = 0;
			while (!isRunning() && nChecks++ <= 20)
				sleep(1);
			if (nChecks <= 20) {
				//	ok - up and receptive; set nice colormap, zoom in a bit
				//	and switch to scale log
				send("orient none");
				send("zoom to 2");
				send("scale log");
				send("colormap BB");

				ALTRETURN;
			}
		} else
			//	is already running
			ALTRETURN;
	}

	errstr<<"Could not launch Ds9. It is either not "
			"installed or inaccessible via the current PATH."
		  <<Sas::error("UnableToLaunchProgram");

	ENDPROC
} // Ds9::Ds9



void Ds9::flipX() const
{
	PROC

	vector<string> v(receive("orient"));

	//
	//	depending on what the current orientation is we need to
	//	do send different commands to achieve the desired result:
	//
	//		current orientation:	new orientation:
	//		--------------------    ----------------
	//		none					x
	//		x						none
	//		y						xy
	//		xy						y
	//
	if (v.size() == 1)
		if (v[0] == "none")
			send("orient x");
		else if (v[0] == "x")
			send("orient none");
		else if (v[0] == "y")
			send("orient xy");
		else if (v[0] == "xy")
			send("orient y");
		else
			errstr<<"Current image orientation is `"<<v[0]<<"\' - "
					"not understood"
				  <<Sas::warning("UnknownOrientation");
	else
		errstr<<"Could not get actual orientation of the image - not flipped"
			  <<Sas::warning("UnknownOrientation");

	ENDPROC
} // Ds9::flipX()


void Ds9::flipY() const
{
	PROC

	vector<string> v(receive("orient"));

	//
	//	depending on what the current orientation is we need to
	//	do send different commands to achieve the desired result:
	//
	//		current orientation:	new orientation:
	//		--------------------    ----------------
	//		none					x
	//		x						none
	//		y						xy
	//		xy						y
	//
	if (v.size() == 1)
		if (v[0] == "none")
			send("orient y");
		else if (v[0] == "x")
			send("orient xy");
		else if (v[0] == "y")
			send("orient none");
		else if (v[0] == "xy")
			send("orient x");
		else
			errstr<<"Current image orientation is `"<<v[0]<<"\' - "
					"not understood"
				  <<Sas::warning("UnknownOrientation");
	else
		errstr<<"Could not get actual orientation of the image - not flipped"
			  <<Sas::warning("UnknownOrientation");

	ENDPROC
} // Ds9::flipY()


void Ds9::zoomIn(ZoomFactor) const
{
	PROC

	errstr<<"Zooming in not yet implemented - nothing done"
		  <<Sas::warning("NotYetImplemented");

	ENDPROC
} // Ds9::zoomIn()


void Ds9::zoomOut(ZoomFactor) const
{
	PROC

	errstr<<"Zooming out not yet implemented - nothing done"
		  <<Sas::warning("NotYetImplemented");

	ENDPROC
} // Ds9::zoomOut()


void Ds9::displayImage(const string &dataSetName) const
{
	PROC

	send("file " + dataSetName);

	ENDPROC
} // Ds9::displayImage()


void Ds9::displayRegion(const string &regionTable) const
{
	PROC

	InASCregionTable inRegionTable(regionTable);
	GeoRegion region;

	inRegionTable>>region;	//	read region from table
	char *regionFile = tmpnam(0);
	ofstream of(regionFile);
	if (!of)
		errstr<<"Cannot open region file `"<<regionFile<<"' for write"
			  <<Sas::error("UnableToCreateFile");

	of<<Ds9::Ds9Region(region, Ds9Region::LINEAR)<<endl;
	of.close();

//	send("regions deleteall");
//	send("zoom to 1");
	send("regions file " + (string)regionFile);

	unlink(regionFile);

	ENDPROC
} // Ds9::displayRegion()


vector<string> Ds9::regionSelection() const
{
	PROC
	//	get raw image coordinates back in SAOtng format
	RETURN receive("regions -coord image -format ds9");
} // Ds9::regionSelection()


void Ds9::down() const
{
	PROC

	send("exit");

	ENDPROC
} // Ds9::down()


Ds9::~Ds9()
{
	PROC
	ENDPROC
} // Ds9::~Ds9()


void Ds9::send(const string &command) const
{
	PROC

	string commandLine = Xpaset + " " + _title;
	FILE *xpasetPipe = popen(commandLine.c_str(), "w");
	if (!xpasetPipe ||
		fprintf(xpasetPipe, "%s\n", command.c_str()) <= 0)
		errstr<<"Could not open or write on pipe to `"<<Xpaset
			  <<"' - please note that `"<<Xpaset<<"' must be accessible via"
				" the PATH. Could this be the problem?"
			  <<Sas::warning("UnableToSendCommand");

	//	see above why whis is necessay
	switchSignalHandlers();
	pclose(xpasetPipe);
	switchSignalHandlers();

	ENDPROC
} // Ds9::send()


vector<string> Ds9::receive(const string &what) const
{
	PROC

	switchSignalHandlers();
	vector<string> response;
	string commandLine = Xpaget + " " + _title + " " + what + " 2>&1";
	FILE *xpagetPipe = popen(commandLine.c_str(), "r");
	if (xpagetPipe) {
		char s[2048];
		while (fgets(s, sizeof(s)-1, xpagetPipe)) {
			if (strlen(s) && s[strlen(s)-1] == '\n')
				s[strlen(s)-1] = '\0';
			if (s[0] != '#')	//	comments
				response.push_back(s);
		}
	} else
		errstr<<"Could not open or write on pipe to `"<<Xpaget
			  <<"' - please note that `"<<Xpaget<<"' must be accessible via"
				" the PATH. Could this be the problem?"
			  <<Sas::warning("UnableToReceive");

	//	see above why whis is necessay
	pclose(xpagetPipe);
	switchSignalHandlers();

	RETURN response;
} // Ds9::receive()


bool Ds9::isRunning() const
{
	PROC

	vector<string> v(receive(""));

	if (v.size() == 0)
		ALTRETURNV(true);

	v = breakup(v[0], " ");

	RETURN v[0] != "XPA$ERROR";
} // Ds9::isRunning()


bool Ds9::isThere() const
{
	PROC
	RETURN true;
} // Ds9::isThere()


void Ds9::execBackground(const string &command) const
{
	PROC

	string comm = command + " >/dev/null 2>&1 &";
	system(comm.c_str());

	ENDPROC
} // Ds9::execBackground()


static ShapeBasics::ShapeColor toColor(const string &colString)
{
	PROC

	ShapeBasics::ShapeColor color = ShapeBasics::BLACK;
	string col(lowercase(colString));
	if (col == "black") color = ShapeBasics::BLACK;
	else if (col == "red") color = ShapeBasics::RED;
	else if (col == "green") color = ShapeBasics::GREEN;
	else if (col == "blue") color = ShapeBasics::BLUE;
	else if (col == "cyan") color = ShapeBasics::CYAN;
	else if (col == "magenta") color = ShapeBasics::MAGENTA;
	else if (col == "yellow") color = ShapeBasics::YELLOW;
	else if (col == "white") color = ShapeBasics::WHITE;

	RETURN color;	
} // toColor()


::GeoRegion Ds9::toRegion(const vector<string> &regionSpec,
						  const string &ignoreMarkerColor)
{
	PROC

	bool ignoreMarkers = !ignoreMarkerColor.empty();
	ShapeBasics::ShapeColor ignCol = ShapeBasics::BLACK;
	if (ignoreMarkers)
		ignCol = toColor(ignoreMarkerColor);

	string knownRegions = "point circle annulus ellipse polygon box";

	::GeoRegion		region;
	region.push_back(Component());
	Component	&comp = region[0];

	for (unsigned n=0; n<regionSpec.size(); ++n) {
		string s(trim(regionSpec[n], true));	//	make a copy

		if (s.empty() || s[0] == '#')
			continue;	//	skip empty + comment lines (starting with '#')

		//	chop off everything after '#'
		string::size_type pos;
		string comment;
		if ((pos = s.find("#")) != string::npos) {
			comment=trim(s.substr(pos+1), true);
			s=trim(s.substr(0, pos), true);
		}

		if (s.length()<=5 || lowercase(s.substr(0, 6)) == "global")
			continue;	//	not interested in global attributes

		bool linear = false;
		if ((pos = s.find(";")) != string::npos) {
			linear = lowercase(trim(s.substr(0, pos))) == "linear";
			s = trim(s.substr(pos+1), true);
		}

		//
		//	now s should contain just the marker
		//
		vector<string> fields(breakup(s, "(, )"));
		if (fields.size() < 3)
			continue;

		#define THROWINVALID(minsize)	\
			if (fields.size() < minsize) { throw GeoRegion::InvalidRegion(); }

		string regionType = lowercase(fields[0]);
		bool included = regionType[0] != '-';
		if (!included)
			regionType.erase(0, 1);

		if (knownRegions.find(regionType) == string::npos)
			continue;

		if (regionType != "polygon") {
			FPoint mid(toNumber<float>(fields[1]),
					   toNumber<float>(fields[2]));
			if (regionType == "point")
				comp.push_back(new FShape(mid));
			else if (regionType == "circle") {
				THROWINVALID(4);
				comp.push_back(new FCircle(mid, toNumber<float>(fields[3])));
			} else if (regionType == "annulus") {
				THROWINVALID(5);
				comp.push_back(new FAnnulus(mid, toNumber<float>(fields[3]),
												 toNumber<float>(fields[4])));
			} else if (regionType == "ellipse") {
				THROWINVALID(6);
				comp.push_back(new FEllipse(mid, toNumber<float>(fields[3]),
												 toNumber<float>(fields[4]),
												 toNumber<float>(fields[5])));
			} else if (regionType == "box") {
				THROWINVALID(6);
				comp.push_back(new FRotBox(mid,	toNumber<float>(fields[3]),
												toNumber<float>(fields[4]),
												toNumber<float>(fields[5])));
			}
		} else { // polygon
			vector<float> x, y;
			unsigned N = fields.size(), i = 1;
			while (i+2 <= N && !isalpha(fields[i][0])) {
				x.push_back(toNumber<float>(fields[i]));
				y.push_back(toNumber<float>(fields[i+1]));
				i += 2;
			}
			comp.push_back(new FPolygon(x, y, x.size()));
		}
		FShape *shape = comp.back();
		shape->include(included);

		//
		//	process optional arguments in comments
		//		color=
		//		text={} | text="bla"	
		//
		vector<string> v(quotedBreakup(comment, "= ", "\"{"));
		if (v.size()<2)
			continue;
		for (unsigned n=0; n<v.size(); n+=2) {
			string key = lowercase(v[n]),
				   arg = v[n+1];
			if (key=="color")
				shape->color(toColor(arg));
			else if (key=="text") {
				string text(unquote(arg));
				if (!text.empty() && text[0] == '{') {
					text.erase(0, 1);
					pos = text.find("}");
					if (pos != string::npos)
						text.erase(pos, 1);
				}
				shape->label(text);
			}
			//	ignore the rest for moment - if we want to process more
			//	options this should happen here
		}

		//
		//	check if we want to ignore this component
		//
		if (ignoreMarkers && shape->color() == ignCol) {
			delete shape;
			comp.pop_back();			
		}
	}	

	RETURN region;
} // Ds9::toRegion


template<class T>
ostream &operator<<(ostream &os, const Point<T> &p)
{
	PROC
	os<<p.x()<<','<<p.y();
	return os;
	ENDPROC
}

ostream &operator<<(ostream &os, const Ds9::Ds9Region &r)
{
	PROC

	static const char* colorString[] = { "black", "white", "red", "green",
										 "blue", "cyan", "magenta", "yellow" };

	os<<"# Region file format: DS9 version 3.0"<<endl;
	os<<"global color=white font=\"helvetica 10 normal\"";
	os<<" edit=1 move=1 delete=1 include=1 fixed=0"<<endl;
	for (unsigned i=0; i<r.size(); ++i) {
		const Component &c = r[i];
		for (unsigned n=0; n<c.size(); ++n) {
			bool commentStarted = false;
			switch (r.coordType()) {
				case Ds9::Ds9Region::FK5:
					os<<"fk5;";
					break;
				case Ds9::Ds9Region::IMAGE:
					os<<"image;";
					break;
				default:
					os << "linear;";
					break;
			}
			if (!c[n]->included())
				os<<'-';
			switch(c[n]->type()) {
				case ShapeBasics::POINT:
					os<<"point("<<c[n]->refPoint()<<") # point=circle";
					commentStarted = true;
					break;
				case ShapeBasics::CIRCLE:
					os<<"circle("<<c[n]->refPoint()<<','<<c[n]->radii()[0]
					  <<')';
					break;
				case ShapeBasics::ELLIPSE:
					os<<"ellipse("<<c[n]->refPoint()<<','<<c[n]->radii()[0]
					  <<','<<c[n]->radii()[1]<<','<<c[n]->rotAngles()[0]<<')';
					break;
				case ShapeBasics::ROTBOX:
					os<<"box("<<c[n]->refPoint()<<','<<c[n]->radii()[0]
					  <<','<<c[n]->radii()[1]<<','<<c[n]->rotAngles()[0]<<')';
					break;
				case ShapeBasics::POLYGON: {
					os<<"polygon(";
					vector<float> x = c[n]->pointsX(),
								  y = c[n]->pointsY();
					os<<x[0]<<','<<y[0];
					for (unsigned n=1; n<x.size(); ++n)
						os<<','<<x[n]<<','<<y[n];
					os<<')';
					break;
				}
				default:
					errstr<<"Shape `"<<c[n]->name()<<"' is not supported by "
							"ds9 - will be skipped"
						  <<Sas::warning("UnsupportedShape");
					break;
			}
			//	write optional attributes
			if (!commentStarted)
				os<<" #";
			os<<" color="<<colorString[UNT(c[n]->color())];
			if (c[n]->isLabeled())
				os<<" text={"<<c[n]->label()<<"}";
			os<<endl;
		}
	}

	RETURN os;
} // operator<<
