//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	Ds9.h:	--- wrapper class around Ds9
//
//	DESCRIPTION:
//		A class `Ds9' is defined which implements the `ImageDisplayer'
//		interface using the image display program Ds9.
//
//	CHANGE HISTORY:
//		UL, 1999-05-06: 1.0.0: - first version
//		UL, 1999-12-15: 1.1.0: - added functionality to convert ds9 region
//								 file string into selectlib Region object
//
//	AUTHOR:
//		U. Lammers (UL), Thu May  6 11:14:42 MET DST 1999
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/task_devel/packages/imgdisplay/src/Ds9.h,v 1.10 2000/10/09 15:46:18 ulammers Exp $
//

#ifndef DS9_H
#define DS9_H

#include <iostream>
#include "ImageDisplayer.h"
#include "Region.h"

class Ds9 : public virtual ImageDisplayer {
public:
	Ds9(const string &name = "ds9");

	class Ds9Region : public GeoRegion {
	public:
		enum CoordType {IMAGE, LINEAR, FK5};
	  
		Ds9Region(const GeoRegion &r, 
				  const CoordType coordType = IMAGE)
			: GeoRegion(r), _coordType(coordType) {}

		CoordType coordType() const { return _coordType; }
	private:
		CoordType _coordType;
	};

	bool isUp() const { PROC RETURN isRunning(); }

	void flipX() const;
	void flipY() const;

	void zoomIn(ZoomFactor zoomFactor = X1) const;
	void zoomOut(ZoomFactor zoomFactor = X1) const;

	void displayImage(const string &dataSetName) const;
	void displayRegion(const string &regionTable) const;
	vector<string> regionSelection() const;

	static ::GeoRegion toRegion(const vector<string> &regionSpec,
								const string &ignoreMarkerColor = "");
		//
		//	PURPOSE:
		//		Convert a string vector with region specifications into
		//		a <GeoRegion> object
		//
		//	ARGUMENTS:
		//		regionSpec: vector with regions specification strings
		//					in Ds9 v3.0 format, e.g.
		//					"-circle(348.5,359,44) # text={yyy/zzz} color=red"
		//		ignoreMarkerColor:
		//					a string that must be a valid Ds9 region marker
		//					color or empty; in the former case all markers
		//					with this color will be ignored, i.e., not
		//					be used in the region construction; this is
		//					to allow for selective import of regions
		//
		//	RETURNS:
		//		GeoRegion object corresponding to given <regionSpec> string;
		//		all elements are packed into one Component
		//

	void send(const string &command) const;
	vector<string> receive(const string &command) const;

	void down() const;

	~Ds9();

private:
	static const string Xpaset;	//	send commands to Ds9
	static const string Xpaget;	//	read settings from Ds9
	static const string TITENV;	//	"Ds9-$LOGNAME"

	//	methods
	bool isRunning() const;
		//	Is Ds9 running?

	bool isThere() const;
		//	Is the executable there?

	void execBackground(const string &command) const;
		//	execute command in the background

	//	data members
	string	_title;	//	window title of Ds9
	string	_DS9;	//	name of the game

}; // Ds9;


ostream &operator<<(ostream &os, const Ds9::Ds9Region &region);
	//
	//	PURPOSE:
	//		Write a region to an ostream in the Ds9 region file format
	//

#endif

